export interface IPublicKey {
  userId: string
  publicKeyArmored: string
  roomId: string
}

export interface IPrivateKey {
  privateKey: any
  passphrase: string
  username: string
  privateKeyArmored: string
  publicKeyArmored: string
  revocationCertificate: string
  roomId: string
}
export interface IMetaMessage {
  whisper?: boolean
}

export interface IWriteMessage {
  clearString: string
  userId: string
  roomId: string
}

export interface IReadMessage {
  encString: string
  meta: IMetaMessage
  roomId: string
}

export class EndToEnd {
  openpgp: any

  keys: {
    publics: IPublicKey[]
    privates: IPrivateKey[]
  } = {
    publics: [],
    privates: []
  }

  onSend: (payload: string, roomId: string, userId: string) => Promise<any>

  constructor(opts: {
    openpgp?: any
    onSend?: (encString: string, roomId: string, userId: string) => Promise<any>
    keys?: {
      publics: IPublicKey[]
      privates: IPrivateKey[]
    }
  }) {
    // if (!opts) throw new Error('no configuration provided')

    if (opts.openpgp) this.openpgp = opts.openpgp
    if (opts.onSend) this.onSend = opts.onSend
    if (opts.keys) this.keys = opts.keys
  }

  addUser(user: IPublicKey) {
    if (!user || !user.publicKeyArmored || !user.userId || !user.roomId) throw new Error('wrong params for adduser')
    this.keys.publics.push(user)
  }

  async whisperToUser(message: IWriteMessage) {
    const msg = await this.composeMessage(message, { whisper: true })

    await this.onSend(msg.encString, message.roomId, message.userId)
  }
  async sendToUsers(message: IWriteMessage, meta) {
    const msgs = await this.composeMessagesForUsers(Object.assign({ meta }, message))

    for (const msg of msgs) {
      await this.onSend(msg.roomId, message.roomId, message.userId)
    }
  }
  async composeMessage(message: IWriteMessage, meta: IMetaMessage, openpgp?: any): Promise<IReadMessage> {
    const key = this.keys.publics.find(f => f.userId === message.userId)
    if (!key) throw new Error(message.userId + " key's not found")
    const mgs = { clearString: message.clearString, keys: this.keys.publics }
    const encMsg = Object.assign({ openpgp }, mgs)
    const encrypted = await this.encryptString(encMsg)
    const m: IReadMessage = {
      encString: encrypted,
      roomId: message.roomId,
      meta: meta
    }
    return m
  }

  async composeMessagesForUsers(opts: { roomId: string; clearString: string; meta }, openpgp?: any) {
    const users = this.keys.publics.filter(f => f.roomId === opts.roomId)

    const messages: IReadMessage[] = []

    for (const user of users) {
      const clearUserMessage: IWriteMessage = {
        roomId: opts.roomId,
        clearString: opts.clearString,
        userId: user.userId
      }
      const encUserMessage = await this.composeMessage(clearUserMessage, opts.meta || {}, openpgp)

      messages.push(encUserMessage)
    }

    return messages
  }

  async readMessage(message: IReadMessage, openpgp?: any) {
    const key = this.keys.privates.find(f => f.roomId === message.roomId)
    const opts = {
      openpgp: openpgp || this.openpgp,
      encString: message.encString,
      privateKey: key.privateKey
    }

    return await this.decryptString(opts)
  }

  async decryptString(opts: { encString: string; privateKey: any; openpgp?: any }) {
    return await EndToEnd.decryptString(this.openpgp || opts.openpgp, opts.privateKey, opts.encString)
  }
  async encryptString(opts: {
    clearString: string
    publicKeyArmored?: string
    userId?: string
    openpgp?: any
  }): Promise<string> {
    if (opts.userId && this.keys.publics && this.keys.publics.length) {
      const key = this.keys.publics.find(f => f.userId === opts.userId)
      if (!key) throw new Error(opts.userId + " key's not found")
      const settings = {
        publicKeyArmored: key.publicKeyArmored,
        clearString: opts.clearString
      }
      return await EndToEnd.encryptString(this.openpgp || opts.openpgp, settings)
    } else if (opts.publicKeyArmored) {
      const settings = {
        publicKeyArmored: opts.publicKeyArmored,
        clearString: opts.clearString
      }

      return await EndToEnd.encryptString(this.openpgp || opts.openpgp, settings)
    } else {
      throw new Error('wrong params')
    }
  }

  async setKey(opts: { opengp?: any; username: string; passphrase: string; roomId: string; userId: string }) {
    const roomKey = this.keys.publics.find(f => f.roomId === opts.roomId)

    if (roomKey) return console.warn('room key just set')

    const keys = await EndToEnd.regenKey(this.openpgp || opts.opengp, {
      username: opts.username,
      passphrase: opts.passphrase
    })

    const newRoomKey = {
      privateKeyArmored: keys.privateKeyArmored,
      publicKeyArmored: keys.publicKeyArmored,
      revocationCertificate: keys.revocationCertificate,
      privateKey: keys.privateKey,
      roomId: opts.roomId,
      userId: opts.userId
    }

    this.keys.publics.push(newRoomKey)
  }

  static async regenKey(openpgp, opts: { username: string; passphrase: string }) {
    if (!openpgp) throw new Error('no opengp provided')
    const opgp = await openpgp.generateKey({
      userIds: [{ name: opts.username }], // you can pass multiple user IDs
      curve: 'ed25519', // ECC curve name
      passphrase: opts.passphrase // protects the private key
    })
    const privateKey = (await openpgp.key.readArmored(opgp.privateKeyArmored)).keys[0]
    await privateKey.decrypt(opts.passphrase)
    const keys = {
      privateKeyArmored: opgp.privateKeyArmored,
      publicKeyArmored: opgp.publicKeyArmored,
      revocationCertificate: opgp.revocationCertificate,
      privateKey: privateKey
    }
    return keys
  }

  static async encryptString(openpgp, opts: { clearString: string; publicKeyArmored: string }) {
    if (!openpgp) throw new Error('no opengp provided')
    const encrypted = await openpgp.encrypt({
      message: openpgp.message.fromText(opts.clearString), // input as Message object
      publicKeys: (await openpgp.key.readArmored(opts.publicKeyArmored)).keys // for encryption
    })
    return encrypted.data
  }

  static async decryptString(openpgp, privateKey, encString: string) {
    if (!openpgp) throw new Error('no opengp provided')
    const { data: decrypted } = await openpgp.decrypt({
      message: await openpgp.message.readArmored(encString), // parse armored message
      privateKeys: [privateKey] // for decryption
    })
    return decrypted
  }
}
